#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import subprocess

REMOTE_URL = 'https://github.com/dmsovetov/Dreemchest'
BRANCH = 'renderer-refactoring'


class Command:
    """A command line tool to checkout a source code from a remote repository"""

    def __init__(self, parser):
        """Constructs a pull command line tool"""

        parser.set_defaults(function=self.checkout)

    @staticmethod
    def checkout(options):
        """Clones a remote repository"""

        command_line = 'git clone %s --branch %s --single-branch xxx' % (REMOTE_URL, BRANCH);

        try:
            subprocess.check_call(command_line, shell=True, stdout=None, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            pass



#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################


class Tool:
    """A base class for all command line tools that have nested commands"""

    def __init__(self, parser, description):
        """Constructs a command line tool instance"""
        self._commands = parser.add_subparsers(help=description)

    def _add_command(self, name, cls):
        """Adds a nested command"""

        cls(self._commands.add_parser(name))


def add_component(parser, name, description=None, options=[]):
    """Adds a library component option to a parser object"""

    option = parser.add_mutually_exclusive_group()
    option.add_argument('--no-%s' % name.lower(),
                        help=description,
                        action='store_true',
                        default=False)

    if len(options) > 1:
        option.add_argument('--renderer',
                            help='specifies a rendering backend to be used.',
                            type=str,
                            choices=options,
                            default=options[0])


def add_system_library(parser, name, versions=None):
    """Adds a system library option to a parser object"""

    option = parser.add_mutually_exclusive_group()
    option.add_argument('--no-%s' % name.lower(),
                        help='disables a Qt library support.',
                        action='store_true',
                        default=False)
    option.add_argument('--%s' % name.lower(),
                        help='specifies the %s library version that is required by a build system.' % name,
                        type=str,
                        choices=versions,
                        default=versions[0])


def add_library(parser, name, is_bundled=True):
    """Adds a third party library option to a parser object"""

    option = parser.add_mutually_exclusive_group()
    option.add_argument('--no-%s' % name.lower(),
                        help='disables the %s library support.' % name,
                        action='store_true',
                        default=False)
    option.add_argument('--system-%s' % name.lower(),
                        help='use %s from the operating system if possible.' % name,
                        action='store_true',
                        default=False if is_bundled else True)

    if is_bundled:
        option.add_argument('--%s' % name.lower(),
                            help='use %s bundled with a source code distribution.' % name,
                            action='store_true',
                            default=True)


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os
import sys
import cmake
import command_line


class PlatformConfigurationCommand(cmake.Command):
    """A base class for all platform configuration commands"""

    def __init__(self, parser, rendering_backend):
        """Constructs a platform configuration command"""

        cmake.Command.__init__(self, parser)

        command_line.add_component(parser, 'pch', description='generate a build system that uses precompiled headers.')
        command_line.add_component(parser, 'composer', description='do not build the Composer tool.')
        command_line.add_component(parser, 'relight', description='do not build the Relight lightmapping library.')
        command_line.add_component(parser, 'tests', description='do not build unit tests.')
        command_line.add_component(parser, 'examples', description='do not build examples.')
        command_line.add_component(parser, 'sound', description='build with no sound support.')
        command_line.add_component(parser, 'renderer', description='build with no rendering support.', options=rendering_backend)

        # Add third party libraries options
        self._libraries = ['tiff', 'jsoncpp', 'zlib', 'Box2D', 'gtest', 'png', 'lua', 'ogg', 'vorbis', 'OpenAL', 'curl']

        for lib in self._libraries:
            command_line.add_library(parser, lib)

        parser.set_defaults(function=self.configure)

    def configure(self, options):
        """Performs basic build system configuration"""
        pass

    def _prepare(self, options):
        """Generates a command line arguments from an input options"""

        # Generate platform independent CMake arguments from passed options
        parameters = dict(
            DC_USE_PCH=cmake.enable_option(options.pch),
            DC_COMPOSER_ENABLED=cmake.disable_option(options.no_composer or options.no_qt),
            DC_OPENGL_ENABLED=cmake.disable_option(options.no_renderer),
            DC_SOUND_ENABLED=cmake.disable_option(options.no_sound),
            DC_WITH_RELIGHT=cmake.disable_option(options.no_relight),
            DC_BUILD_TESTS=cmake.disable_option(options.no_tests),
            DC_BUILD_EXAMPLES=cmake.disable_option(options.no_examples),
            CMAKE_CXX_STANDARD=options.cpp,
            DC_QT_SUPPORT=('disabled' if options.no_qt else options.qt).capitalize(),
            PREFIX_PATH=os.path.abspath(options.prefix_path),
        )

        # Generate CMake arguments from added libraries
        for lib in self._libraries:
            cmake.library_option(parameters, lib, options)

        return parameters


class DesktopConfigureCommand(PlatformConfigurationCommand):
    """Performs desktop OS build system configuration"""

    def __init__(self, parser, rendering_backend):
        """Constructs desktop configuration command"""

        PlatformConfigurationCommand.__init__(self, parser, rendering_backend)

        # Add Qt library option
        command_line.add_system_library(parser, 'Qt', versions=['auto', 'qt4', 'qt5'])

        # Add third party libraries
        self._libraries.append('FBX')
        command_line.add_library(parser, 'FBX', is_bundled=False)


class WindowsConfigureCommand(DesktopConfigureCommand):
    """Performs Windows build system configuration"""

    def __init__(self, parser):
        """Constructs Windows configuration command"""

        DesktopConfigureCommand.__init__(self, parser, ['opengl', 'direct3d9', 'direct3d12'])

    def configure(self, options):
        """Performs basic build system configuration"""

        # Perform basic build configuration
        cmake_parameters = self._prepare(options)

        # Invoke a CMake command to generate a build system
        cmake.generate('Visual Studio 12', options.source, options.output, cmake_parameters)


class MacOSConfigureCommand(DesktopConfigureCommand):
    """Performs MacOS build system configuration"""

    def __init__(self, parser):
        """Constructs MacOS configuration command"""

        DesktopConfigureCommand.__init__(self, parser, ['opengl'])

    def configure(self, options):
        """Performs MacOS build system configuration"""

        # Perform basic build configuration
        cmake_parameters = self._prepare(options)

        # Invoke a CMake command to generate a build system
        cmake.generate('Xcode', options.source, options.output, cmake_parameters)


class Command(command_line.Tool):
    """A command line tool to configure a build system for a specified target system"""

    def __init__(self, parser):
        """Constructs a configure command line tool"""

        command_line.Tool.__init__(self, parser, 'available platforms')

        if os.name == 'nt':
            self._add_command('windows', WindowsConfigureCommand)
        if sys.platform == 'darwin':
            self._add_command('macos', MacOSConfigureCommand)


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import cmake
import shutil
import os
import cmake
import command_line


class InstallCommand(cmake.Command):
    """A base class for all installation commands"""

    def __init__(self, parser, generators=[]):
        """Constructs a third party install command"""

        cmake.Command.__init__(self, parser, output_dir='Build/Dependencies')

        self._add_library(parser, 'zlib')
        self._add_library(parser, 'Box2D')
        self._add_library(parser, 'OpenAL')
        self._add_library(parser, 'GTest')
        self._add_library(parser, 'jsoncpp')
        self._add_library(parser, 'libpng')
        self._add_library(parser, 'libtiff')
        self._add_library(parser, 'lua')
        self._add_library(parser, 'ogg')
        self._add_library(parser, 'vorbis')
        self._add_library(parser, 'curl')

        parser.add_argument('--generator',
                            help='build system generator to be used.',
                            choices=generators,
                            default=generators[0])

        parser.set_defaults(function=self.install)

    @staticmethod
    def install(options):
        install_path = os.path.abspath(options.output)
        binary_dir = os.path.abspath(os.path.join(options.source, 'Projects'))
        source_dir = os.path.abspath(options.source)

        if not options.no_zlib:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'zlib'),
                                      os.path.join(binary_dir, 'zlib'),
                                      prefix=install_path,
                                      target='install'
                                      )

        if not options.no_curl:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'curl'),
                                      os.path.join(binary_dir, 'curl'),
                                      prefix=install_path,
                                      options=dict(BUILD_CURL_EXE=False, CURL_STATICLIB=True),
                                      target='install'
                                      )

        if not options.no_box2d:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'Box2D/Box2D'),
                                      os.path.join(binary_dir, 'Box2D'),
                                      prefix=install_path,
                                      options=dict(BOX2D_BUILD_EXAMPLES=False, BOX2D_INSTALL=True),
                                      target='install'
                                      )

        if not options.no_jsoncpp:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'jsoncpp'),
                                      os.path.join(binary_dir, 'jsoncpp'),
                                      prefix=install_path,
                                      options=dict(JSONCPP_WITH_TESTS=False, JSONCPP_WITH_POST_BUILD_UNITTEST=False),
                                      target='install'
                                      )

        if not options.no_ogg:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'ogg'),
                                      os.path.join(binary_dir, 'ogg'),
                                      prefix=install_path,
                                      target='install'
                                      )

        if not options.no_vorbis:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'vorbis'),
                                      os.path.join(binary_dir, 'vorbis'),
                                      prefix=install_path,
                                      target='install'
                                      )

        if not options.no_gtest:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'GoogleTest'),
                                      os.path.join(binary_dir, 'GoogleTest'),
                                      prefix=install_path,
                                      target='install'
                                      )

        if not options.no_libpng:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'libpng'),
                                      os.path.join(binary_dir, 'libpng'),
                                      prefix=install_path,
                                      options=dict(PNG_TESTS=False),
                                      target='install'
                                      )

        if not options.no_libtiff:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'libtiff'),
                                      os.path.join(binary_dir, 'libtiff'),
                                      prefix=install_path,
                                      options=dict(CMAKE_CXX_FLAGS='-D_XKEYCHECK_H'),
                                      target='install'
                                      )

        if not options.no_lua:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'lua'),
                                      os.path.join(binary_dir, 'lua'),
                                      prefix=install_path,
                                      options=dict(BUILD_SHARED_LIBS=False),
                                      target='install'
                                      )

        if not options.no_openal:
            cmake.configure_and_build(options.generator,
                                      os.path.join(source_dir, 'OpenAL'),
                                      os.path.join(binary_dir, 'OpenAL'),
                                      prefix=install_path,
                                      options=dict(ALSOFT_EXAMPLES=False, ALSOFT_TESTS=False, LIBTYPE='STATIC'),
                                      target='install'
                                      )

        shutil.rmtree(binary_dir)

    @staticmethod
    def _add_library(parser, name):
        """Adds a third party build options"""
        parser.add_argument('--no-%s' % name.lower(),
                            help='do not build a %s.' % name,
                            action='store_true',
                            default=False
                            )


class MacOSInstallCommand(InstallCommand):
    """A command line tool to build and install third party libraries bundled with a source code distribution"""

    def __init__(self, parser):
        """Constructs a third party install command"""

        InstallCommand.__init__(self, parser, generators=['Xcode'])


class WindowsInstallCommand(InstallCommand):
    """A command line tool to build and install third party libraries bundled with a source code distribution"""

    def __init__(self, parser):
        """Constructs a third party install command"""

        InstallCommand.__init__(self, parser, generators=['Visual Studio 12'])


class Command(command_line.Tool):
    """A command line tool to build and install dependencies for a specified target system"""

    def __init__(self, parser):
        """Constructs a configure command line tool"""

        command_line.Tool.__init__(self, parser, 'available platforms')

        self._add_command('windows', WindowsInstallCommand)
        self._add_command('macos', MacOSInstallCommand)


#################################################################################
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Dmitry Sovetov
#
# https://github.com/dmsovetov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os
import subprocess
import shutil


def enable_option(value):
    """Converts a boolean option to a CMake ON/OFF switch"""
    return 'ON' if value else 'OFF'


def disable_option(value):
    """Converts a boolean option to a CMake ON/OFF switch"""
    return 'OFF' if value else 'ON'


def library_option(parameters, name, options):
    """Adds a library parameter base on passed options"""

    # Convert arguments to dictionary
    values = vars(options)

    # Convert a library name to lowercase
    name = name.lower()

    if values['no_%s' % name]:
        return

    if values['system_%s' % name]:
        parameters['SYSTEM_%s' % name.upper()] = True
    else:
        parameters['BUNDLED_%s' % name.upper()] = True


def generate(generator, source, output, parameters, rest='', quiet=False):
    """Invokes a CMake command to generate a build system"""

    # Create an output directory
    if not os.path.exists(output):
        os.makedirs(output)

    # Generate CMake parameter string
    cmd_args = ['-D%s=%s' % (key, parameters[key]) for key in parameters]

    # Generate a command line string
    command_line = 'cmake -E chdir %s cmake %s -G "%s" %s %s' % (output, os.path.abspath(source),
                                                                 generator,
                                                                 ' '.join(cmd_args),
                                                                 rest
                                                                 )

    with open(os.devnull, 'wb') as devnull:
        subprocess.check_call(command_line, shell=True, stdout=devnull if quiet else None, stderr=subprocess.STDOUT)


def build(source, target, configuration, quiet=False):
    """Build a CMake-generated project binary tree."""

    # Generate a command line string
    command_line = 'cmake --build %s --target %s --config %s' % (source, target, configuration)

    with open(os.devnull, 'wb') as devnull:
        subprocess.check_call(command_line, shell=True, stdout=devnull if quiet else None, stderr=subprocess.STDOUT)


def configure_and_build(generator, source_dir, binary_dir,
                        target='ALL_BUILD',
                        configuration='Release',
                        prefix=None,
                        options=None):
    """Generates a binary tree and then builds a specified target for requested configuration"""

    if not options:
        options = dict()

    # Set the CMAKE_INSTALL_PREFIX variable
    if prefix:
        options['CMAKE_INSTALL_PREFIX:PATH'] = prefix

    try:
        # Generate a binary tree
        generate(generator, source_dir, binary_dir, options, '-Wno-dev')

        # Build a generated tree
        build(binary_dir, target, configuration)

    except subprocess.CalledProcessError:
        print 'Failed to build %s' % source_dir

    shutil.rmtree(binary_dir)


class Command:
    """A base class for all CMake commands"""

    def __init__(self, parser, source_dir='.', output_dir='.'):
        """Constructs a platform configuration command"""

        parser.add_argument('--source',
                            help='a source directory to generate build system.',
                            default=source_dir)
        parser.add_argument('--output',
                            help='an output directory to place generated build system.',
                            default=output_dir)
        parser.add_argument('--prefix-path',
                            help='a path used for searching packages.',
                            default='.')
        parser.add_argument('--cpp',
                            help='specifies the C++ standard whose features that are required by a build system.',
                            type=int,
                            choices=[98, 11, 14],
                            default=98)

        parser.set_defaults(function=self.configure)

    def configure(self, options):
        """Performs basic build system configuration"""
        pass
